////////////////////////////////////////////////////////////
// AUTOMATION CODE
//////////////////////////////////////////////////////////// 
// This code was added by William Hooper www.whooper.co.uk 
//  to allow automation inside the XLL.
//////////////////////////////////////////////////////////// 
// Remove this file from the project if you do not need automation
//////////////////////////////////////////////////////////// 

#include "stdafx.h"
#include "stdlib.h"
#include "xlcall.h"
#include "framewrk.h"
#include "windows.h"

///////////////////////////////////////////////////////////////////
// Global variables 
///////////////////////////////////////////////////////////////////

LPDISPATCH pExcelApp=NULL;
LPDISPATCH pExcelWorkbooks=NULL;

/////////////////////////////////////////////////////////////////
// Function Prototypes
/////////////////////////////////////////////////////////////////

int GetDispatch( LPDISPATCH &m_pDocDisp, char* s);
HRESULT AutomationWrap(int autoType, VARIANT *pvResult, IDispatch *pDisp, LPOLESTR ptName, int cArgs...);

///////////////////////////////////////////////////////////////////
// Setup Automation
//  Sets pExcelApp and pExcelWorkbooks 
//   Creates a temp workbook and finds it's IDispatch
///////////////////////////////////////////////////////////////////

void AutomationSetup( ) 
{ 

// Create a new workbook with a vb module 
//  and save it to the temp directory
XLOPER xTemp, xRandNum;
unsigned char sTemp[255];

XLOPER xFalse;
xFalse.xltype = xltypeBool;
xFalse.val.boolie = FALSE;

XLOPER xTrue;
xTrue.xltype = xltypeBool;
xTrue.val.boolie = TRUE;

XLOPER xTypeNum;
xTypeNum.xltype = xltypeNum;
xTypeNum.val.num = 6;

// Turn screen updating off so user does not see all this
Excel4(xlcEcho, 0, 1, &xFalse);
// Open New workbook with Visual basic module to initialize OLE
Excel4(xlcNew, 0, 1, &xTypeNum);

// Build xTemp & sTemp file name
xTemp.xltype=xltypeStr;
xTemp.val.str=sTemp;
GetTempPath(245,(char*) &sTemp[1]);
strcat((char*) &sTemp[1], "tmpxll");
sTemp[0] = strlen((char*) &sTemp[1]);
Excel4(xlfRand, &xRandNum, 0 );   // returns a xltypeNum
sTemp[sTemp[0]+1] = 97 + int( 26 * xRandNum.val.num);
Excel4(xlfRand, &xRandNum, 0 );   // returns a xltypeNum
sTemp[sTemp[0]+2] = 97 + int(26 * xRandNum.val.num);
sTemp[sTemp[0]+3]=0;
strcat((char*) &sTemp[1], ".xls");
sTemp[0]= sTemp[0]+6;
//Excel4( xlcAlert, 0, 1 , &xTemp);

// Save book 
Excel4(xlcSaveAs , 0, 1, &xTemp);

LPDISPATCH pDocDisp;
if ( !GetDispatch( pDocDisp, (char*) &sTemp[ sTemp[0]-11 ] ) )
{
	// Get Excel Application IDispatch
	VARIANT v1;
	VariantInit(&v1);
	AutomationWrap(DISPATCH_PROPERTYGET, &v1, pDocDisp, L"Parent", 0);
	pExcelApp = v1.pdispVal;
	pDocDisp->Release();

	// Get Workbooks collection
	VARIANT v2;
	VariantInit(&v2);
	AutomationWrap(DISPATCH_PROPERTYGET, &v2, pExcelApp, L"Workbooks", 0);
	pExcelWorkbooks = v2.pdispVal;
}
else
{
	XLOPER xMessage, xInt3;
	xInt3.xltype = xltypeInt;
	xInt3.val.w = 3;
	xMessage.xltype= xltypeStr;
	xMessage.val.str = (PUCHAR) "\34XLL Automation Setup Failed!"; 
	Excel4( xlcAlert, 0, 2 , &xMessage, &xInt3);
}

// Close temp book, delete file, turn screen updating on
Excel4(xlcFileClose,0,0);
DeleteFile((char*) &sTemp[1]);
Excel4(xlcEcho, 0, 1, &xTrue);

return ;
}

///////////////////////////////////////////////////////////////////
// Terminate Automation 
///////////////////////////////////////////////////////////////////

void AutomationTerminate()
{
if (pExcelApp!=NULL) pExcelApp->Release();
if (pExcelWorkbooks!=NULL) pExcelWorkbooks->Release();
return;
}

///////////////////////////////////////////////////////////////////
// Automation helper function 
///////////////////////////////////////////////////////////////////

HRESULT AutomationWrap(int autoType, VARIANT *pvResult, IDispatch *pDisp, LPOLESTR ptName, int cArgs...) {
    // Begin variable-argument list...
    va_list marker;
    va_start(marker, cArgs);

    if(!pDisp) {
        MessageBox(NULL, "NULL IDispatch passed to AutoWrap()", "Error", 0x10010);
        _exit(0);
    }

    // Variables used...
    DISPPARAMS dp = { NULL, NULL, 0, 0 };
    DISPID dispidNamed = DISPID_PROPERTYPUT;
    DISPID dispID;
    HRESULT hr;
    char buf[200];
    char szName[200];

    
    // Convert down to ANSI
    WideCharToMultiByte(CP_ACP, 0, ptName, -1, szName, 256, NULL, NULL);
    
    // Get DISPID for name passed...
    hr = pDisp->GetIDsOfNames(IID_NULL, &ptName, 1, LOCALE_USER_DEFAULT, &dispID);
    if(FAILED(hr)) {
        sprintf(buf, "IDispatch::GetIDsOfNames(\"%s\") failed w/err 0x%08lx", szName, hr);
        MessageBox(NULL, buf, "AutoWrap()", 0x10010);
        _exit(0);
        return hr;
    }
    
    // Allocate memory for arguments...
    VARIANT *pArgs = new VARIANT[cArgs+1];
    // Extract arguments...
    for(int i=0; i<cArgs; i++) {
        pArgs[i] = va_arg(marker, VARIANT);
    }
    
    // Build DISPPARAMS
    dp.cArgs = cArgs;
    dp.rgvarg = pArgs;
    
    // Handle special-case for property-puts!
    if(autoType & DISPATCH_PROPERTYPUT) {
        dp.cNamedArgs = 1;
        dp.rgdispidNamedArgs = &dispidNamed;
    }
    
    // Make the call!
    hr = pDisp->Invoke(dispID, IID_NULL, LOCALE_SYSTEM_DEFAULT, autoType, &dp, pvResult, NULL, NULL);
    if(FAILED(hr)) {
        sprintf(buf, "IDispatch::Invoke(\"%s\"=%08lx) failed w/err 0x%08lx", szName, dispID, hr);
        MessageBox(NULL, buf, "AutoWrap()", 0x10010);
        _exit(0);
        return hr;
    }
    // End variable-argument section...
    va_end(marker);
    
    delete [] pArgs;
    
    return hr;
}

///////////////////////////////////////////////////////////////////
// Get the iDispatch of a given twelve character workbook name
// return 0 if failed or 1 if ok
///////////////////////////////////////////////////////////////////

int GetDispatch( LPDISPATCH &m_pDocDisp, char* s)
{
m_pDocDisp=NULL;

// Get a BindCtx.
IBindCtx *pbc;
HRESULT hr = CreateBindCtx(0, &pbc);
if(FAILED(hr)) {
  // failed
  return 1;
}

// Get running-object table.
IRunningObjectTable *prot;
hr = pbc->GetRunningObjectTable(&prot);
if(FAILED(hr)) {
  // failed
  pbc->Release();
  return 1;
}

// Get enumeration interface.
IEnumMoniker *pem;
hr = prot->EnumRunning(&pem);
if(FAILED(hr)) {
  // failed
  prot->Release();
  pbc->Release();
  return 1;
}

// Start at the beginning.
pem->Reset();

// Churn through enumeration.
ULONG fetched;
IMoniker *pmon;
int n = 0;
while(pem->Next(1, &pmon, &fetched) == S_OK) {

  // Get DisplayName.
  LPOLESTR pName;
  pmon->GetDisplayName(pbc, NULL, &pName);
  // Convert it to ASCII.
  char szName[512];
  WideCharToMultiByte(CP_ACP, 0, pName, -1, szName, 512, NULL,
  NULL);

  // Compare last eight byes with sTemp
  if( !memcmp( &szName[strlen(szName)-12] , s, 12 ) ) {

      // Bind to this ROT entry.
      IDispatch *pDisp;
      hr = pmon->BindToObject(pbc, NULL, IID_IDispatch, (void
      **)&pDisp);
      if(!FAILED(hr)) {
          m_pDocDisp = pDisp;
      }
      else {
          return 1;
      }
  }

  // Release interfaces.
  pmon->Release();

  // Break out if we obtained the IDispatch successfully.
  if(m_pDocDisp != NULL) break;
}

// Release interfaces.
pem->Release();
prot->Release();
pbc->Release();

if(m_pDocDisp == NULL) return 1;


return 0;
}
