////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
// Includes
////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "stdlib.h"

#include "xlcall.h"
#include "framewrk.h"

/////////////////////////////////////////////////////////////////
// External
/////////////////////////////////////////////////////////////////

extern HRESULT AutomationWrap(int autoType, VARIANT *pvResult, IDispatch *pDisp, LPOLESTR ptName, int cArgs...);
extern LPDISPATCH pExcelApp;
extern LPDISPATCH pExcelWorkbooks;

/////////////////////////////////////////////////////////////////
// Function Prototypes
/////////////////////////////////////////////////////////////////

void excelstr_copy( PUCHAR sTo, PUCHAR sFrom);
void excelstr_copynullterm( PUCHAR sTo, char* sFrom);

////////////////////////////////////////////////////////////////////
// Global variables -----------------------------------------------
///////////////////////////////////////////////////////////////////

int iCounter;

////////////////////////////////////////////////////////////////////
// Auto_Open Module ------------------------------------------------
//////////////////////////////////////////////////////////////////

void Auto_Open()    // gets called once on xll open
{
iCounter=0;
return ;
}

////////////////////////////////////////////////////////////////
// Exported Functions & Macros
////////////////////////////////////////////////////////////////


// First worksheet function example : Junk1
// Adds two numbers passed as doubles and returns a double
double __stdcall Junk1( double d1, double d2) 
{
return d1+d2;
}

// Second worksheet function example : Junk2
// Multiplies two numbers passed as OPERs and returns a pointer to a double
// x1 and x2 are defined as OPERs ("P") in the REGISTER data type arguments
// We return a pointer so the return result is held in a static variable.
//  If we used a local variable it would be trashed on return before
//   excel had a chance to use copy it.
double* __stdcall Junk2( LPXLOPER x1, LPXLOPER x2) 
{
static double dResult;

if ( x1->xltype==xltypeNum && x2->xltype==xltypeNum)
{
	// x1 and x2 are numbers
	dResult = x1->val.num * x2->val.num;
	return &dResult;
}
else
{
	//x1 and x2 are not both numbers
	return 0; // returns #NUM! to excel
}
}

// Third worksheet function example : Junk3
// Returns global variable.
// Open two copies of Excel and try this function.
// Each copy loads and runs an independent copy of the XLL
int __stdcall Junk3( LPXLOPER x) 
{
return ++iCounter;
}

// Command Macro Example
// Use Application.Run to call from VBA
// Accepts and returns an integer
int __stdcall CommandMacro( int i ) 
{
	HMENU hMenu = CreateMenu();
	MENUITEMINFO mii;
	memset(&mii, 0, sizeof(MENUITEMINFO));
	mii.cbSize = sizeof(MENUITEMINFO);
    mii.fMask = MIIM_FTYPE | MIIM_STRING | MIIM_STATE | MIIM_ID;
    mii.dwTypeData = "blah";
    mii.cch = 6;
    mii.dwItemData = 0;
    mii.wID = 1;
    mii.fState = 0;
    BOOL b = InsertMenuItem(hMenu, 0, TRUE, &mii);
	//BOOL b  = InsertMenuItem(hMenu, 1, TRUE, &mii);
	static HWND hWnd = NULL;
	HMENU hMenuOld = GetMenu(hWnd);
	b = SetMenu(hWnd, hMenu);

	HWND xxx=NULL;
	b = ShowWindow(xxx,SW_HIDE);


XLOPER xStr; 
const UCHAR s[]="\14Hello\nWorld"; // notice octal byte count 
xStr.xltype= xltypeStr;
xStr.val.str = (PUCHAR) s;

Excel4( xlcAlert, 0, 1 , &xStr);

/*
//Alternatives:

	Excel( xlcAlert, 0, 1 , TempStr(" Hello\nWold"));

// or

	Excel4( xlcAlert, 0, 1 , TempStr(" Hello\nWold"));
	FreeAllTempMemory();

// or

	UCHAR su[255];
	excelstr_copynullterm(su, "Hello\nWold");
	xStr.val.str = su;
	Excel4( xlcAlert, 0, 1 , &xStr);
*/

return i+1;
}

// Command Macro Ole Example
void __stdcall CommandMacroOle( ) 
{
XLOPER xStr; 
const UCHAR s[]="\16Ole Example...";  
xStr.xltype= xltypeStr;
xStr.val.str = (PUCHAR) s;
Excel4( xlcAlert, 0, 1 , &xStr);


// Call Workbooks.Add() to get a new workbook...
IDispatch *pXlBook;
{
  VARIANT result;
  VariantInit(&result);
  AutomationWrap(DISPATCH_PROPERTYGET, &result, pExcelWorkbooks, L"Add", 0);
  pXlBook = result.pdispVal;
}

// Create a 15x15 safearray of variants...
VARIANT arr;
arr.vt = VT_ARRAY | VT_VARIANT;
{
  SAFEARRAYBOUND sab[2];
  sab[0].lLbound = 1; sab[0].cElements = 15;
  sab[1].lLbound = 1; sab[1].cElements = 15;
  arr.parray = SafeArrayCreate(VT_VARIANT, 2, sab);
}

// Fill safearray with some values...
for(int i=1; i<=15; i++) {
  for(int j=1; j<=15; j++) {
     // Create entry value for (i,j)
     VARIANT tmp;
     tmp.vt = VT_I4;
     tmp.lVal = i*j;
     // Add to safearray...
     long indices[] = {i,j};
     SafeArrayPutElement(arr.parray, indices, (void *)&tmp);
  }
}

// Get ActiveSheet object
IDispatch *pXlSheet;
{
  VARIANT result;
  VariantInit(&result);
  AutomationWrap(DISPATCH_PROPERTYGET, &result, pExcelApp, L"ActiveSheet", 0);
  pXlSheet = result.pdispVal;
}

// Get Range object for the Range A1:O15...
IDispatch *pXlRange;
{
  VARIANT parm;
  parm.vt = VT_BSTR;
  parm.bstrVal = ::SysAllocString(L"A1:O15");

  VARIANT result;
  VariantInit(&result);
  AutomationWrap(DISPATCH_PROPERTYGET, &result, pXlSheet, L"Range", 1, parm);
  VariantClear(&parm);

  pXlRange = result.pdispVal;
}

// Set range with our safearray...
AutomationWrap(DISPATCH_PROPERTYPUT, NULL, pXlRange, L"Value", 1, arr);


// Set .Saved property of workbook to TRUE so we aren't prompted
// to save when we tell Excel to quit...
{
  VARIANT x;
  x.vt = VT_I4;
  x.lVal = 1;
  AutomationWrap(DISPATCH_PROPERTYPUT, NULL, pXlBook, L"Saved", 1, x);
}


// Release references...
pXlRange->Release();
pXlSheet->Release();
pXlBook->Release();
VariantClear(&arr);

return;
}

//////////////////////////////////////////////////////
// Function by William Hooper www.whooper.co.uk
// Returns the contents of the calling cell
// Notice # in the register data type argument to avoid circular ref message
//////////////////////////////////////////////////////
LPXLOPER __stdcall ThisCell( )
{
static XLOPER xResult;
static UCHAR sResult[255];

XLOPER xValue, xThisCell, xNum; 

Excel4(xlfCaller, &xThisCell, 0);  // Returns an xltypeSRef

// xNum holds an integer specifying the types to coerce to
xNum.xltype = xltypeInt;
xNum.val.w = xltypeNum | xltypeStr | xltypeErr | xltypeBool;

Excel(xlCoerce, &xValue, 2, &xThisCell, &xNum); // Coerce to a value

if (xValue.xltype==xltypeStr)
{
	excelstr_copy(sResult, xValue.val.str);
	xResult.xltype=xltypeStr;
	xResult.val.str=sResult;
	Excel4(xlFree,0,1,&xValue);
}
else
{
	xResult=xValue;
}

return &xResult;
}

//////////////////////////////////////////////////////////////
// Function by William Hooper www.whooper.co.uk
// This function is equivalent to NOW() but is non volatile
/////////////////////////////////////////////////////////////
LPXLOPER __stdcall NowNonVolatile( LPXLOPER xDummyDependency) 
{

static XLOPER xRes;

Excel4(xlfNow, &xRes, 0); // Returns an xltypeNum;

return &xRes;

}

//////////////////////////////////////////////////////////////
// Function by William Hooper www.whooper.co.uk
// 2d Lookup function
/////////////////////////////////////////////////////////////
LPXLOPER __stdcall Lookup2d( LPXLOPER xRange, LPXLOPER xRow, LPXLOPER xCol ) 
{
static XLOPER xResult;
static XLMREF xlmrefResultRef;
XLOPER xMatch,xIndex, xInt1, xInt0;
int iRow, iCol;
if (xRange->xltype!=xltypeSRef && xRange->xltype!=xltypeRef)
{
	xResult.xltype=xltypeErr;
	xResult.val.err=xlerrName;
	return &xResult;
}
xInt1.xltype=xltypeInt;
xInt1.val.w=1;
xInt0.xltype=xltypeInt;
xInt0.val.w=0;
Excel4( xlfIndex, &xIndex, 3, xRange, &xInt0, &xInt1);  // first column
Excel4( xlfMatch, &xMatch, 3, xRow, &xIndex, &xInt0);   // lookup row
Excel4(xlFree,0,1,&xIndex);  // we need to free xIndex - it's probably an xltypeRef
if (xMatch.xltype!=xltypeNum)  // but not xMatch as it's only a num or error (no attached data)
{	// Not found - return #REF!
	xResult.xltype=xltypeErr;
	xResult.val.err=xlerrRef;
	return &xResult;
}
iRow= (int) xMatch.val.num;
Excel4( xlfIndex, &xIndex, 3, xRange, &xInt1, &xInt0);  // first row
Excel4( xlfMatch, &xMatch, 3, xCol, &xIndex, &xInt0);   // lookup col
Excel4(xlFree,0,1,&xIndex);
if (xMatch.xltype!=xltypeNum)
{	// Not found - return #REF!
	xResult.xltype=xltypeErr;
	xResult.val.err=xlerrRef;
	return &xResult;
}
iCol=(int) xMatch.val.num;
if (xRange->xltype==xltypeRef)
{	// xltypeRef
	xResult.xltype= xltypeRef;
	xResult.val.mref.idSheet=xRange->val.mref.idSheet;
	xResult.val.mref.lpmref=&xlmrefResultRef;
	xlmrefResultRef.count=1;
	xlmrefResultRef.reftbl[0].rwFirst=xRange->val.mref.lpmref->reftbl[0].rwFirst + iRow-1;
	xlmrefResultRef.reftbl[0].rwLast=xlmrefResultRef.reftbl[0].rwFirst;
	xlmrefResultRef.reftbl[0].colFirst=xRange->val.mref.lpmref->reftbl[0].colFirst + iCol-1;
	xlmrefResultRef.reftbl[0].colLast=xlmrefResultRef.reftbl[0].colFirst;
	return &xResult;
}
else
{	// xltypeSRef
	xResult.xltype= xltypeSRef;
	xResult.val.sref.count=1;
	xResult.val.sref.ref.rwFirst = xRange->val.sref.ref.rwFirst + iRow-1;
	xResult.val.sref.ref.rwLast = xResult.val.sref.ref.rwFirst;
	xResult.val.sref.ref.colFirst = xRange->val.sref.ref.colFirst + iCol-1;
	xResult.val.sref.ref.colLast= xResult.val.sref.ref.colFirst;
	return &xResult;
}
}


////////////////////////////////////////////////////////////////
// Private Functions
////////////////////////////////////////////////////////////////

// Copy s2 into s1
// s1 and s2 are excel style byte counted strings
void excelstr_copy( PUCHAR sTo, PUCHAR sFrom)
{
sTo[0] = sFrom[0];
for (int i=1 ; i<=sFrom[0]; i++) 
	sTo[i] = sFrom[i];
return;
}

// Copy Null termminated sFrom into excel style sTo
// Checks max length of 255 is not exceeded
void excelstr_copynullterm( PUCHAR sTo, char* sFrom)
{
int i=0;
while (i<255 && sFrom[i])
	sTo[i+1]=sFrom[i++];
sTo[0]=i;
return;
}

////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////